# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
import filecmp
import locale
import os

from dateutil import tz


class MediaFactory(object):

    def __init__(self, *factories):
        self.factories = factories

    def __call__(self, path):
        return next((
            f(path) for f in self.factories
            if path.lower().endswith(f.file_extensions)), None)


class FileSystemElement(object):

    def __init__(self, path):
        pref_enc = locale.getpreferredencoding()
        if isinstance(path, unicode):
            self._path = path
        else:
            self._path = path.decode(pref_enc)

    def move_to(self, path_prefix):
        """
        Move the file to a new path prefix.  The file extension must not be
        part of path_prefix as it is added by this method using the
        file_extension method.
        :param path_prefix: the destination path without extension
        """
        path = get_destination_path(
            source=self.path(),
            dest=path_prefix,
            extension=self.file_extension())
        if not os.path.exists(path):
            os.renames(self._path, path)
        elif self._path != path:
            os.remove(self._path)
        self._path = path

    def link_to(self, path_prefix):
        """
        Link the file to a new path prefix.  The file extension must not be
        part of path_prefix as it is added by this method using the
        file_extension method.
        :param path_prefix: the destination path without extension
        """
        dest_link_path = get_destination_path(
            source=self.path(),
            dest=path_prefix,
            extension=self.file_extension())
        if not os.path.exists(dest_link_path):
            parent, _ = os.path.split(path_prefix)
            if parent and not os.path.exists(parent):
                os.makedirs(parent)
            try:
                os.link(self._path, dest_link_path)
            except os.error:
                relative_src_path = os.path.join(
                    os.path.relpath(
                        os.path.dirname(self._path),
                        os.path.dirname(dest_link_path)),
                    os.path.basename(self._path))
                os.symlink(relative_src_path, dest_link_path)

    def path(self):
        """
        Return file path (including file extension)
        """
        return self._path

    def file_extension(self):
        """
        Return the actual or required file extension.
        """
        _, file_extension = os.path.splitext(self.path())
        return file_extension


def file_exist_as_symlink(path):
    """Test whether a path exists.  Returns False for broken symbolic links"""
    try:
        os.stat(path)
    except os.error:
        return False
    return True



class Location(object):
    pass


# noinspection PyAbstractClass
class Media(FileSystemElement):
    file_extensions = ()
    time_zone = tz.gettz()

    def location(self):
        """
        :rtype: Location
        """
        raise NotImplemented()

    def timestamp(self):
        """
        Return the creation timestamp of the media.  The name is parsed to be
        used as a date/time.  If it is not a date, the modification time of
        the file on the file system is used.
        :rtype: datetime
        """
        try:
            _, file_name = os.path.split(self.path())
            return string_to_datetime(file_name, self.time_zone)
        except ValueError:
            return datetime.fromtimestamp(
                timestamp=os.path.getmtime(self.path()),
                tz=self.time_zone
            )

    def file_extension(self):
        """
        Return the required file extension
        """
        if len(self.file_extensions):
            return self.file_extensions[0]
        return super(Media, self).file_extension()


# noinspection PyAbstractClass
class MediaCollection(FileSystemElement):
    media_factory = Media

    def __iter__(self):
        for sub_path, subdirs, files in os.walk(self.path()):
            for name in files:
                media = self.media_factory(os.path.join(sub_path, name))
                if media is not None:
                    yield media


# noinspection PyAbstractClass
class TimeLine(MediaCollection):
    """
    A media collection organized by date.
    """
    media_path_format = "%Y/%Y-%m/%Y-%m-%d-%H-%M-%S"

    def link(self, media):
        """
        Link the media file in this TimeLine MediaCollection.
        :param media: the media file
        """
        new_path = os.path.join(
            self.path(),
            media.timestamp().strftime(self.media_path_format))
        media.link_to(new_path)

    def move(self, media):
        """
        Move the media file in this TimeLine MediaCollection.
        :param media: the media file
        """
        new_path = os.path.join(
            self.path(),
            media.timestamp().strftime(self.media_path_format))
        media.move_to(new_path)

    def organize(self):
        """
        Reorganize he files in this TimeLine MediaCollection using the
        configured media path format.
        """
        for media in self:
            self.move(media)


# noinspection PyAbstractClass
class Album(FileSystemElement):
    """
    A file element created by an event that contains a timeline media
    collection.
    """
    timeline_factory = TimeLine

    def title(self):
        """
        Returns the title of the album.
        """
        _, name = os.path.split(self.path())
        return name

    def locations(self):
        """
        :rtype: Location
        """
        raise NotImplementedError()

    def timeline(self):
        """
        Return the media collection organized by date.
        :rtype: TimeLine
        """
        if not hasattr(self, '_timeline'):
            self._timeline = self.timeline_factory(path=self.path())
        return self._timeline

    @classmethod
    def from_event(cls, event, root_path):
        """
        Create the album in the specified root_path using the provided event.
        :rtype: Album
        """
        file_safe_title = ''.join(
            c for c in event.title() if c not in '\/:*?<>|').strip(' .')
        return cls(os.path.join(root_path, file_safe_title))


class TimePeriod(object):
    """
    Represents a block of time with a start date/time and a duration that may
    be recurrent or not.
    """

    def start(self):
        """
        Get the beginning or the first occurrence of this time period.
        :return: start of this time period as a datetime.
        :rtype: datetime.datetime
        """
        raise NotImplementedError()

    def duration(self):
        """
        Get the duration of the first occurrence of this time period.
        :return: duration of the first occurrence of this time period.
        :rtype: datetime.timedelta
        """
        raise NotImplementedError()

    def __contains__(self, timestamp):
        """
        Checks if the time period includes a specific timestamp.
        :param timestamp: a timezone aware datetime
        :return: if the timestamp is included
        """
        raise NotImplementedError()


class Event(object):
    """
    Represents a calendar event with a title and a time period.
    """

    def title(self):
        """
        Returns the title of the event.
        """
        raise NotImplementedError()

    def time_period(self):
        """
        Returns the time period of this event.
        :rtype TimePeriod
        """
        raise NotImplementedError()

    def location(self):
        """
        :rtype Location
        """
        raise NotImplementedError()

    def save_to(self, folder):
        """
        Save this event in the provided folder.
        :param folder: The folder where the event will be saved
        """
        raise NotImplementedError()


def is_same_file(source, dest):
    return os.path.samefile(source, dest) or \
              filecmp.cmp(source, dest, shallow=False)


def get_destination_path(source, dest, extension, suffix=0):
    if suffix:
        computed_dest = u'{}({}){}'.format(dest, suffix, extension)
    else:
        computed_dest = u'{}{}'.format(dest, extension)
    if os.path.exists(computed_dest):
        if is_same_file(source, computed_dest):
            return computed_dest
        return get_destination_path(
            source=source, dest=dest, extension=extension, suffix=suffix+1)
    return computed_dest


def string_to_datetime(string, tzinfo):
    date_string = ''.join(c for c in string if c.isdigit())[0:14]
    if len(date_string) != 14:
        raise ValueError('time data in "{}" must contain 14 digits'.format(string))
    dt = datetime.strptime(date_string, '%Y%m%d%H%M%S')
    return dt.replace(tzinfo=tzinfo)



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import hashlib
import shutil
import tempfile

import yaml


def file_path(file_name):
    return os.path.join(os.path.dirname(__file__), file_name)


def copytree():
    tmp_folder = tempfile.mkdtemp()
    inbox_path = os.path.join(tmp_folder, 'inbox')
    shutil.copytree(os.path.dirname(__file__), inbox_path)
    return tmp_folder, inbox_path


def md5sum(filename):
    with open(filename,"rb") as fd:
        return hashlib.md5(fd.read()).hexdigest()


files = yaml.safe_load(open(file_path('images.yaml')))



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import os
import unittest

from dateutil import tz
from hamcrest import assert_that, is_
from calbum.core.model import TimeLine

from calbum.sources.image import JpegPicture
from tests import resources


class TestExifPicture(unittest.TestCase):
    def test_exif(self):
        assert_that(
            JpegPicture(resources.file_path('image-01.jpeg')).timestamp(),
            is_(datetime(2012, 5, 1, 1, 0, 0, tzinfo=tz.gettz())))


class TestExifTimeLine(unittest.TestCase):
    def test_timeline_organize_pictures(self):
        _, inbox_path = resources.copytree()
        timeline = TimeLine(inbox_path)

        timeline.organize()

        for name, file_details in resources.files.iteritems():
            file_path = os.path.join(inbox_path, file_details['expected_path'])
            assert_that(
                os.path.exists(file_path),
                is_(True),
                'File is missing: {} -> {}'.format(name, file_path))
            assert_that(
                resources.md5sum(file_path),
                is_(file_details['md5sum']),
                'wrong md5sum for {} at {}'.format(name, file_path)
            )




# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
from dateutil.tz import gettz

from progress.bar import ChargingBar

from calbum.core import model
from calbum.filters import timeline, album, NoopMediaFilter
from calbum.sources import image, calendar, exiftool

model.MediaCollection.media_factory = model.MediaFactory(
    image.JpegPicture,
    image.TiffPicture,
    exiftool.VideoMP4Media,
    exiftool.Video3GPMedia,
)

def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        prog='calbum',
        add_help=True,
        description='calbum is an unattended calendar-based photo organiser. '
                    'It is meant to allow easy management of pictures based '
                    'on their location, date and calendar events without '
                    'the need of a database or a special browser to retrieve '
                    'them.')

    parser.add_argument('--link-only',
                        help='Keep files where they are in the inbox folder.',
                        action='store_true')

    parser.add_argument('--inbox',
                        help='The path of the inbox directory. '
                             '(default: ./inbox)',
                        metavar='path',
                        default='./inbox')

    parser.add_argument('--timeline',
                        help='The path of the timeline directory. '
                             '(default: ./timeline)',
                        metavar='path',
                        default='./timeline')

    parser.add_argument('--album',
                        help='The path of the album directory. '
                             '(default: ./album)',
                        metavar='path',
                        default='./album')

    parser.add_argument('--calendar',
                        help='The url of the album calendar. (ical)',
                        metavar='url')

    parser.add_argument('--date-format',
                        help='The format to use for timestamps.',
                        metavar='format',
                        default=model.TimeLine.media_path_format)

    parser.add_argument('--save-events',
                        help='Keep the calendar event in the album.',
                        action='store_true')

    parser.add_argument('--time-zone',
                        metavar='tz',
                        help='Pictures timezone (default to local time).')

    settings = vars(parser.parse_args(args))

    # Configure data model
    model.TimeLine.media_path_format = settings['date_format']
    model.Media.time_zone = gettz(settings['time_zone'])

    # Create filters
    timeline_filter = timeline.TimelineFilter(settings['timeline'])
    album_filter = NoopMediaFilter()
    if settings['calendar']:
        events = calendar.CalendarEvent.load_from_url(url=settings['calendar'])
        album_filter = album.CalendarAlbumFilter(
            albums_path=settings['album'],
            events=events,
            save_events=settings['save_events']
        )

    filter_actions = [
        timeline_filter.link if settings['link_only'] else timeline_filter.move,
        album_filter.link
    ]

    # Perform actions
    suffix = '%(index)d/%(max)d [eta: %(eta)ds]'
    bar = ChargingBar('Processing inbox:', suffix=suffix)
    for picture in bar.iter(list(model.MediaCollection(settings['inbox']))):
        for action in filter_actions:
            action(picture)



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import unittest

from dateutil import tz
from hamcrest import assert_that, is_, instance_of
from calbum.sources import exiftool

from tests import resources


class TestExifToolMedia(unittest.TestCase):
    def setUp(self):
        self.exiftool_path = exiftool.exiftool_path

    def tearDown(self):
        exiftool.exiftool_path = self.exiftool_path

    def test_jpeg_timestamp(self):
        assert_that(
            exiftool.JpegPicture(resources.file_path('image-01.jpeg')).timestamp(),
            is_(datetime(2012, 5, 1, 1, 0, 0, tzinfo=tz.gettz())))

    def test_tiff_timestamp(self):
        assert_that(
            exiftool.TiffPicture(resources.file_path('image-03.tif')).timestamp(),
            is_(datetime(2013, 2, 1, 3, 0, 0, tzinfo=tz.gettz())))

    def test_mp4_timestamp(self):
        assert_that(
            exiftool.VideoMP4Media(resources.file_path('video-01.mp4')).timestamp(),
            is_(datetime(2014, 1, 1, 19, 30, 0, tzinfo=tz.gettz())))

    def test_3gp_timestamp(self):
        assert_that(
            exiftool.Video3GPMedia(resources.file_path('video-02.3gp')).timestamp(),
            is_(datetime(2014, 2, 2, 19, 30, 0, tzinfo=tz.gettz())))

    def test_mp4_timestamp_without_exiftool(self):
        try:
            exiftool.ExifToolMedia.exiftool_path = '__invalid_command__'
            assert_that(
                exiftool.VideoMP4Media(resources.file_path('video-01.mp4')).timestamp(),
                instance_of(datetime))
        except Exception as e:
            raise AssertionError('Expected no exception but raised {}'.format(e))



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import unittest

from dateutil import tz
from hamcrest import assert_that, is_
import mock

from calbum.core import model


class TestFileSystemElement(unittest.TestCase):

    @mock.patch('os.renames')
    @mock.patch('os.remove')
    @mock.patch('os.path.exists')
    @mock.patch('calbum.core.model.get_destination_path')
    def test_move_to(self, get_destination_path, exists, remove, renames):
        get_destination_path.return_value = 'dest/path.jpg'
        exists.return_value = False

        fse = model.FileSystemElement('origin/path.jpg')
        fse.move_to('dest/path')
        assert_that(fse.path(), is_('dest/path.jpg'))

        get_destination_path.assert_called_with(
            source='origin/path.jpg', dest='dest/path', extension='.jpg')
        exists.assert_called_with('dest/path.jpg')
        renames.assert_called_with('origin/path.jpg', 'dest/path.jpg')
        assert_that(remove.called, is_(False))

    @mock.patch('os.renames')
    @mock.patch('os.remove')
    @mock.patch('os.path.exists')
    @mock.patch('calbum.core.model.get_destination_path')
    def test_move_to_same_file(self, get_destination_path, exists, remove, renames):
        get_destination_path.return_value = 'dest/path.jpg'
        exists.return_value = True

        fse = model.FileSystemElement('origin/path.jpg')
        fse.move_to('dest/path')
        assert_that(fse.path(), is_('dest/path.jpg'))

        get_destination_path.assert_called_with(
            source='origin/path.jpg', dest='dest/path', extension='.jpg')
        exists.assert_called_with('dest/path.jpg')
        remove.assert_called_with('origin/path.jpg')
        assert_that(renames.called, is_(False))

    @mock.patch('os.link')
    @mock.patch('os.makedirs')
    @mock.patch('os.remove')
    @mock.patch('os.path.exists')
    @mock.patch('calbum.core.model.get_destination_path')
    def test_link_to(self, get_destination_path, exists, remove, makedirs, link):
        get_destination_path.return_value = 'dest/path.jpg'
        exists.return_value = False

        fse = model.FileSystemElement('origin/path.jpg')
        fse.link_to('dest/path')
        assert_that(fse.path(), is_('origin/path.jpg'))

        get_destination_path.assert_called_with(
            source='origin/path.jpg', dest='dest/path', extension='.jpg')
        exists.assert_has_calls([
           mock.call('dest/path.jpg'),
           mock.call('dest'),
        ])
        makedirs.assert_called_with('dest')
        link.assert_called_with('origin/path.jpg', 'dest/path.jpg')
        assert_that(remove.called, is_(False))

    @mock.patch('os.link')
    @mock.patch('os.remove')
    @mock.patch('os.path.exists')
    @mock.patch('calbum.core.model.get_destination_path')
    def test_link_to_same_file(self, get_destination_path, exists, remove, link):
        get_destination_path.return_value = 'dest/path.jpg'
        exists.return_value = True

        fse = model.FileSystemElement('origin/path.jpg')
        fse.link_to('dest/path')
        assert_that(fse.path(), is_('origin/path.jpg'))

        get_destination_path.assert_called_with(
            source='origin/path.jpg', dest='dest/path', extension='.jpg')
        exists.assert_called_with('dest/path.jpg')
        assert_that(link.called, is_(False))
        assert_that(remove.called, is_(False))

    @mock.patch('os.symlink')
    @mock.patch('os.link')
    @mock.patch('os.makedirs')
    @mock.patch('os.remove')
    @mock.patch('os.path.exists')
    @mock.patch('calbum.core.model.get_destination_path')
    def test_link_to_make_symlink_on_error(self, get_destination_path, exists, remove, makedirs, link, symlink):
        get_destination_path.return_value = 'dest/new_name.jpg'
        exists.return_value = False
        link.side_effect = OSError()

        fse = model.FileSystemElement('\xc3\x80 trier/origin/path.jpg')
        fse.link_to('dest/new_name')
        assert_that(fse.path(), is_(u'\xc0 trier/origin/path.jpg'))

        get_destination_path.assert_called_with(
            source=u'\xc0 trier/origin/path.jpg', dest='dest/new_name', extension='.jpg')
        exists.assert_has_calls([
            mock.call('dest/new_name.jpg'),
            mock.call('dest'),
            ])
        makedirs.assert_called_with('dest')
        link.assert_called_with(u'\xc0 trier/origin/path.jpg', 'dest/new_name.jpg')
        symlink.assert_called_with(u'../\xc0 trier/origin/path.jpg', 'dest/new_name.jpg')
        assert_that(remove.called, is_(False))

    def test_file_extension_based_on_path(self):
        fse = model.FileSystemElement('some/file/path.pdf')
        assert_that(fse.file_extension(), is_('.pdf'))


class TestGetDestinationPath(unittest.TestCase):

    @mock.patch('os.path.exists')
    def test_get_destination_path_file_doesnt_exist(self, exists):
        exists.return_value = False
        assert_that(
            model.get_destination_path(
                source='origin/path.jpg', dest='dest/path', extension='.jpg'),
            is_('dest/path.jpg'))
        exists.assert_called_with('dest/path.jpg')

    @mock.patch('filecmp.cmp')
    @mock.patch('os.path.samefile')
    @mock.patch('os.path.exists')
    def test_get_destination_path_file_exist_with_different_file(self, exists, samefile, filecmp):
        samefile.return_value = False
        filecmp.return_value = False
        exists.side_effect = [
            True,
            False
        ]
        assert_that(
            model.get_destination_path(
                source='origin/path.jpg', dest='dest/path', extension='.jpg'),
            is_('dest/path(1).jpg'))
        exists.has_calls([
            mock.call('dest/path.jpg'),
            mock.call('dest/path(1).jpg'),
        ])

    @mock.patch('filecmp.cmp')
    @mock.patch('os.path.samefile')
    @mock.patch('os.path.exists')
    def test_get_destination_path_file_exist_with_different_file_twice(self, exists, samefile, filecmp):
        samefile.return_value = False
        filecmp.return_value = False
        exists.side_effect = [
            True,
            True,
            False
        ]
        assert_that(
            model.get_destination_path(
                source='origin/path.jpg', dest='dest/path', extension='.jpg'),
            is_('dest/path(2).jpg'))
        exists.has_calls([
            mock.call('dest/path.jpg'),
            mock.call('dest/path(1).jpg'),
            mock.call('dest/path(2).jpg'),
        ])

    @mock.patch('filecmp.cmp')
    @mock.patch('os.path.samefile')
    @mock.patch('os.path.exists')
    def test_get_destination_path_file_exist_with_same_file_inode(self, exists, samefile, filecmp):
        samefile.return_value = True
        filecmp.return_value = False
        exists.side_effect = [
            True
        ]
        assert_that(
            model.get_destination_path(
                source='origin/path.jpg', dest='dest/path', extension='.jpg'),
            is_('dest/path.jpg'))
        exists.has_calls([
            mock.call('dest/path.jpg'),
        ])

    @mock.patch('filecmp.cmp')
    @mock.patch('os.path.samefile')
    @mock.patch('os.path.exists')
    def test_get_destination_path_file_exist_with_same_file_content(self, exists, samefile, filecmp):
        samefile.return_value = False
        filecmp.return_value = True
        exists.side_effect = [
            True
        ]
        assert_that(
            model.get_destination_path(
                source='origin/path.jpg', dest='dest/path', extension='.jpg'),
            is_('dest/path.jpg'))
        exists.has_calls([
            mock.call('dest/path.jpg'),
        ])


class TestMedia(unittest.TestCase):

    def test_file_extension_based_on_path(self):
        fse = model.Media('some/file/path.pdf')
        assert_that(fse.file_extension(), is_('.pdf'))

    def test_file_extension_based_on_class(self):
        class FakeGifMedia(model.Media):
            file_extensions = ('.gif', '.gifa')
        fse = FakeGifMedia('some/file/path.pdf')
        assert_that(fse.file_extension(), is_('.gif'))

    def test_timestamp_based_on_filename(self):
        media = model.Media('some/file/VID_20120501_224323.avi')
        assert_that(
            media.timestamp(),
            is_(datetime(2012, 5, 1, 22, 43, 23, tzinfo=tz.gettz())))



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import exifread

from calbum.core.model import Media, string_to_datetime


class ExifPicture(Media):
    file_extensions = ()

    timestamp_tags = (
        'Image DateTime',
        'EXIF DateTimeOriginal',
        'EXIF DateTimeDigitized',
        'DateTime'
    )

    def exif(self):
        if not hasattr(self, '_exif'):
            with open(self.path()) as f:
                self._exif = exifread.process_file(f)
        return self._exif

    def timestamp(self):
        """
        Return the creation timestamp of the media as defined in the EXIF
        metadata ('Image DateTime', 'EXIF DateTimeOriginal',
        'EXIF DateTimeDigitized', 'DateTime')
        :rtype: datetime
        """
        exif = self.exif()

        d = next((
            string_to_datetime(str(exif[tag]), self.time_zone)
            for tag in self.timestamp_tags if tag in exif), None)
        if d and (d.tzinfo is None or d.tzinfo.utcoffset(d) is None):
            d = d.replace(tzinfo=self.time_zone)

        return d or super(ExifPicture, self).timestamp()

    def location(self):
        raise NotImplemented()


class JpegPicture(ExifPicture):
    file_extensions = ('.jpeg', '.jpg')


class TiffPicture(ExifPicture):
    file_extensions = ('.tiff', '.tif')



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from calbum.core import model
from calbum.filters import MediaFilter


class CalendarAlbumFilter(MediaFilter):

    def __init__(self, albums_path, events, save_events):
        self.events = list(events)
        self.albums_path = albums_path
        self.save_events = save_events
 
    def albums_for(self, media):
        for event in self.events:
            if media.timestamp() in event.time_period():
                yield (model.Album.from_event(event, self.albums_path),
                       event)

    def move(self, media):
        album, event = next(self.albums_for(media), (None, None))
        if album:
            album.timeline().move(media)
            if self.save_events:
                event.save_to(album.path())

    def link(self, media):
        for album, event in self.albums_for(media):
            album.timeline().link(media)
            if self.save_events:
                event.save_to(album.path())




# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from datetime import datetime, date, timedelta

from hamcrest import assert_that, is_
from icalendar import Event
import pytz

from calbum.sources.calendar import CalendarTimePeriod


class TestEventTimePeriod(unittest.TestCase):

    def test_timestamp_in_simple_event(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART:20151010T100000Z
DTEND:20151010T120000Z
CREATED:20151010T100000Z
UID:123456
END:VEVENT"""))

        assert_that(etp.start(), is_(datetime(2015, 10, 10, 10, tzinfo=pytz.utc)))
        assert_that(etp.duration(), is_(timedelta(hours=2)))

        assert_that(datetime(2015, 10, 10, 11, tzinfo=pytz.utc) in etp, is_(True))

        assert_that(datetime(2015, 10, 10, 9, tzinfo=pytz.utc) in etp, is_(False))

        assert_that(datetime(2015, 10, 10, 13, tzinfo=pytz.utc) in etp, is_(False))

    def test_timestamp_in_simple_recursive_event(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART:20151010T100000Z
DTEND:20151010T120000Z
CREATED:20151010T100000Z
RRULE:FREQ=DAILY;COUNT=2
UID:123456
END:VEVENT"""))

        assert_that(etp.start(), is_(datetime(2015, 10, 10, 10, tzinfo=pytz.utc)))
        assert_that(etp.duration(), is_(timedelta(hours=2)))

        assert_that(datetime(2015, 10, 10, 11, tzinfo=pytz.utc) in etp, is_(True))
        assert_that(datetime(2015, 10, 10, 9, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 10, 10, 13, tzinfo=pytz.utc) in etp, is_(False))


        assert_that(datetime(2015, 10, 11, 11, tzinfo=pytz.utc) in etp, is_(True))
        assert_that(datetime(2015, 10, 11, 9, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 10, 11, 13, tzinfo=pytz.utc) in etp, is_(False))

        assert_that(datetime(2015, 10, 12, 11, tzinfo=pytz.utc) in etp, is_(False))

    def test_timestamp_in_recursive_event_with_exceptions(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART:20151010T100000Z
DTEND:20151010T120000Z
CREATED:20151010T100000Z
RRULE:FREQ=DAILY;COUNT=5
EXDATE:20151011T100000Z,20151012T100000Z
EXDATE:20151013T100000Z
UID:123456
END:VEVENT"""))

        assert_that(etp.start(), is_(datetime(2015, 10, 10, 10, tzinfo=pytz.utc)))
        assert_that(etp.duration(), is_(timedelta(hours=2)))

        assert_that(datetime(2015, 10, 10, 11, tzinfo=pytz.utc) in etp, is_(True))
        assert_that(datetime(2015, 10, 10, 9, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 10, 10, 13, tzinfo=pytz.utc) in etp, is_(False))

        assert_that(datetime(2015, 10, 11, 11, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 10, 12, 11, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 10, 13, 11, tzinfo=pytz.utc) in etp, is_(False))

        assert_that(datetime(2015, 10, 14, 11, tzinfo=pytz.utc) in etp, is_(True))

        assert_that(datetime(2015, 10, 15, 11, tzinfo=pytz.utc) in etp, is_(False))

    def test_timestamp_in_timeless_event(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART;VALUE=DATE:20151113
DTEND;VALUE=DATE:20151114
DTSTAMP:20150925T034508Z
UID:9E783D57-BF34-47B0-86D8-FE6FD83B05A2
SEQUENCE:0
CREATED:20150922T191411Z
LAST-MODIFIED:20150922T191823Z
END:VEVENT"""))

        assert_that(etp.start(), is_(date(2015, 11, 13)))
        assert_that(etp.duration(), is_(timedelta(days=1)))

        assert_that(datetime(2015, 11, 12, 12, tzinfo=pytz.utc) in etp, is_(False))
        assert_that(datetime(2015, 11, 13, 12, tzinfo=pytz.utc) in etp, is_(True))
        assert_that(datetime(2015, 11, 14, 9, tzinfo=pytz.utc) in etp, is_(False))

    def test_timestamp_in_byday_event(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART;TZID=America/Toronto:20150602T183000
DTEND;TZID=America/Toronto:20150602T193000
DTSTAMP:20150925T035639Z
SEQUENCE:0
EXDATE;TZID=America/Toronto:20150827T183000
CREATED:20150519T235325Z
LAST-MODIFIED:20150706T153006Z
RRULE:FREQ=WEEKLY;UNTIL=20150827T223000Z;BYDAY=TU,TH;WKST=SU
END:VEVENT"""))

        assert_that(etp.start(), is_(datetime(2015, 06, 02, 18, 30, tzinfo=pytz.timezone('Etc/GMT+4'))))
        assert_that(etp.duration(), is_(timedelta(hours=1)))

        assert_that(datetime(2015, 11, 12, 12, tzinfo=pytz.timezone('Etc/GMT+4')) in etp, is_(False))

    def test_timestamp_in_yearly_event(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART;VALUE=DATE:20120702
DTEND;VALUE=DATE:20120703
DTSTAMP:20150925T043431Z
SEQUENCE:2
CREATED:20110922T001534Z
LAST-MODIFIED:20150706T153006Z
RRULE:FREQ=YEARLY
END:VEVENT"""))

        assert_that(etp.start(), is_(date(2012, 07, 02)))
        assert_that(etp.duration(), is_(timedelta(days=1)))

        assert_that(datetime(2012, 07, 02, 12) in etp, is_(True))

    def test_timestamp_in_event_without_an_end(self):
        etp = CalendarTimePeriod(Event.from_ical("""BEGIN:VEVENT
DTSTART:20130921T230000Z
DTSTAMP:20150925T051339Z
SEQUENCE:1
CREATED:20130903T183447Z
LAST-MODIFIED:20130916T165029Z
END:VEVENT"""))

        assert_that(etp.start(), is_(datetime(2013, 9, 21, 23, tzinfo=pytz.utc)))
        assert_that(datetime(2013, 9, 21, 23, tzinfo=pytz.utc) in etp, is_(False))



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from hamcrest import assert_that, is_

from calbum import cmd
from calbum.core.model import MediaCollection
from tests import resources


class TestMain(unittest.TestCase):

    def test_main_moving(self):
        repo_path, inbox_path = resources.copytree()
        timeline_path = os.path.join(repo_path, 'timeline')
        album_path = os.path.join(repo_path, 'album')

        os.chdir(repo_path)
        cmd.main([
            '--inbox', inbox_path,
            '--calendar', 'https://raw.githubusercontent.com/JoProvost/calbum/master/tests/resources/calendar.ics',
            '--save-events'
        ])

        timeline = MediaCollection(timeline_path)
        pictures = list(timeline)

        assert_that(len(pictures), is_(len(resources.files)))

        for name, file_details in resources.files.iteritems():
            file_path = os.path.join(
                timeline_path, file_details['expected_path'])
            assert_that(
                os.path.exists(file_path),
                is_(True),
                'File is missing: {} -> {}'.format(name, file_path))
            assert_that(
                resources.md5sum(file_path),
                is_(file_details['md5sum']),
                'wrong md5sum for {} at {}'.format(name, file_path)
            )

            if file_details['event_name']:
                file_path = os.path.join(
                    album_path, file_details['event_name'],
                    file_details['expected_path'])
                assert_that(
                    resources.md5sum(file_path),
                    is_(file_details['md5sum']),
                    'wrong md5sum for {} at {}'.format(name, file_path)
                )

            assert_that(
                os.path.exists(os.path.join(inbox_path, name)),
                is_(False),
                'File is present in inbox: {}'.format(name))

    def test_main_linking(self):
        repo_path, inbox_path = resources.copytree()
        timeline_path = os.path.join(repo_path, 'timeline')
        album_path = os.path.join(repo_path, 'album')

        os.chdir(repo_path)
        cmd.main([
            '--inbox', inbox_path,
            '--calendar', 'https://raw.githubusercontent.com/JoProvost/calbum/master/tests/resources/calendar.ics',
            '--save-events',
            '--link-only',
        ])

        timeline = MediaCollection(timeline_path)
        pictures = list(timeline)

        assert_that(len(pictures), is_(len(resources.files)))

        for name, file_details in resources.files.iteritems():
            file_path = os.path.join(
                timeline_path, file_details['expected_path'])
            assert_that(
                os.path.exists(file_path),
                is_(True),
                'File is missing: {} -> {}'.format(name, file_path))
            assert_that(
                resources.md5sum(file_path),
                is_(file_details['md5sum']),
                'wrong md5sum for {} at {}'.format(name, file_path)
            )

            if file_details['event_name']:
                file_path = os.path.join(
                    album_path, file_details['event_name'],
                    file_details['expected_path'])
                assert_that(
                    resources.md5sum(file_path),
                    is_(file_details['md5sum']),
                    'wrong md5sum for {} at {}'.format(name, file_path)
                )

            assert_that(
                os.path.exists(os.path.join(inbox_path, name)),
                is_(True),
                'File is absent from inbox: {}'.format(name))



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import subprocess

from calbum.core.model import Media, string_to_datetime

exiftool_path = 'exiftool'


class ExifToolMedia(Media):
    file_extensions = ()

    timestamp_tags = (
        'Creation Date',
        'Date/Time Original',
        'Track Create Date',
        'Media Create Date',
        'Create Date',
    )

    def exif(self):
        if not hasattr(self, '_exif'):
            self._exif = {}
            try:
                for line in subprocess.check_output(
                        [exiftool_path, self.path()]).splitlines():
                    key, value = line.split(':', 1)
                    self._exif[key.strip()] = value.strip()
            except OSError as e:
                logging.warning('Metadata processing with "{}" '
                                'failed for "{}": {}'.format(
                    exiftool_path, self.path(), repr(e)))
            except subprocess.CalledProcessError as e:
                logging.warning('Metadata processing with "{}" '
                                'failed for "{}": {}'.format(
                    exiftool_path, self.path(), repr(e)))
        return self._exif

    def timestamp(self):
        """
        Return the creation timestamp of the media as defined in the EXIF
        metadata ('Creation Date', 'Date/Time Original', 'Track Create Date',
        'Media Create Date', 'Create Date')
        :rtype: datetime
        """
        try:
            exif = self.exif()
            d = next((
                string_to_datetime(str(exif[tag]), self.time_zone)
                for tag in self.timestamp_tags if tag in exif), None)
            if d:
                if d.tzinfo is None or d.tzinfo.utcoffset(d) is None:
                    d = d.replace(tzinfo=self.time_zone)
                if d.year < 1970:
                    logging.info('Patched incorrect time offset, for {} '
                                 'see https://trac.ffmpeg.org/ticket/1471'
                                 .format(self.path()))
                    # See https://trac.ffmpeg.org/ticket/1471
                    d = d.replace(year=d.year+66)
                return d
        except ValueError:
            pass

        return super(ExifToolMedia, self).timestamp()

    def location(self):
        raise NotImplemented()


class JpegPicture(ExifToolMedia):
    file_extensions = ('.jpeg', '.jpg')


class TiffPicture(ExifToolMedia):
    file_extensions = ('.tiff', '.tif')


class VideoMP4Media(ExifToolMedia):
    file_extensions = ('.mp4',)


class Video3GPMedia(ExifToolMedia):
    file_extensions = ('.3gp', '.3g2')



# Copyright 2015 Jonathan Provost.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import urllib2
import datetime

import dateutil
import dateutil.rrule
import icalendar

from calbum.core.model import TimePeriod, Event


class CalendarTimePeriod(TimePeriod):
    """
    Represents an icalendar vevent block of time with a start date/time and a
    duration that may be recurrent or not.
    """

    default_duration = datetime.timedelta()

    def __init__(self, event):
        """
        :type event: icalendar.Event
        """
        self._event = event
        self._recurrence = None

    def start(self):
        """
        Get the beginning or the first occurrence of this time period.
        :return: start of this time period as a datetime.
        :rtype: datetime.datetime
        """
        return self._event['dtstart'].dt

    def duration(self):
        """
        Get the duration of the first occurrence of this time period.
        :return: duration of the first occurrence of this time period.
        :rtype: datetime.timedelta
        """
        duration = self._event.get('duration')
        end = self._event.get('dtend')
        if duration:
            return duration.dt
        if end:
            return end.dt - self.start()
        return self.default_duration

    def recurrence(self):
        """
        Get the recurrence rules (with the exceptions).
        :return: the recurrence rules
        :rtype: dateutil.rrule.rruleset
        """
        if self._recurrence is None:
            if 'rrule' in self._event:
                self._recurrence = dateutil.rrule.rruleset()
                self._recurrence.rrule(dateutil.rrule.rrulestr(
                    self._event['rrule'].to_ical(), dtstart=self.start()))
                for dt in get_datetime_list(self._event.get('exdate', [])):
                    self._recurrence.exdate(dt)
        return self._recurrence

    def __contains__(self, timestamp):
        """
        Checks if the time period includes a specific timestamp.
        :param timestamp: a timezone aware datetime
        :return: if the timestamp is included
        """
        start = self.start()
        duration = self.duration()
        rule = self.recurrence()

        if not hasattr(start, 'tzinfo'):
            timestamp = timestamp.replace(tzinfo=None)

        if rule is not None:
            start = rule.before(timestamp, True) or start

        if not hasattr(start, 'tzinfo'):
            timestamp = datetime.date(timestamp.year, timestamp.month, timestamp.day)

        return start <= timestamp < (start+duration)


class CalendarEvent(Event):
    """
    Represents an ical calendar event with a title and a time period.
    """

    def __init__(self, event):
        """
        :type event: icalendar.Event
        """
        self._event = event
        self._time_period = CalendarTimePeriod(self._event)

    def title(self):
        """
        Returns the title of the event (uses the summary field).
        """
        return self._event['summary']

    def time_period(self):
        """
        Returns the time period of this event.
        :rtype TimePeriod
        """
        return self._time_period

    def location(self):
        raise NotImplementedError()

    def save_to(self, folder):
        """
        Save this event in the provided folder (event.ics).
        :param folder: The folder where the event will be saved
        """
        event_path = os.path.join(folder, 'event.ics')
        with open(event_path, 'w') as f:
            cal = icalendar.Calendar()
            cal.add_component(self._event)
            f.write(cal.to_ical())

    @classmethod
    def load_from_file(cls, path):
        with open(path, 'r') as f:
            return cls.load_from_stream(f)

    @classmethod
    def load_from_url(cls, url):
        response = urllib2.urlopen(url)
        if response.code != 200:
            raise Exception("Something went wrong. HTTP response code: %s" % response.code)
        return cls.load_from_stream(response)

    @classmethod
    def load_from_stream(cls, stream):
        events = []
        cal = icalendar.Calendar.from_ical(stream.read().decode('UTF-8'))
        for component in cal.walk():
            if component.name == "VEVENT":
                events.append(cls(event=component))
        return events


def get_datetime_list(obj):
    if not isinstance(obj, list):
        obj = [obj]
    for item in obj:
        if hasattr(item, 'dts'):
            for dt in item.dts:
                yield dt.dt
        else:
            yield item.dt
